import tomllib
from pathlib import Path
from typing import Any, Dict, Optional

from packaging.requirements import Requirement
from packaging.version import Version
from pyproject_metadata import StandardMetadata

from .hwh_config import HwhConfig


class PyProject:
    def __init__(self, project_dir: Path):
        self.project_dir = project_dir
        self.pyproject_path = project_dir / "pyproject.toml"
        self._data: Optional[Dict[str, Any]] = None

    @property
    def toml(self):
        if self._data:
            return self._data

        if not self.pyproject_path.exists():
            raise FileNotFoundError(f"No pyproject.toml found in {self.project_dir}")

        with open(self.pyproject_path, "rb") as f:
            self._data = tomllib.load(f)

        return self._data

    @property
    def metadata(self) -> StandardMetadata:
        return StandardMetadata.from_pyproject(self.toml)

    @property
    def dependencies(self) -> set[Requirement]:
        metadata = self.metadata

        # Combine runtime and build dependencies
        all_deps = set(metadata.dependencies)

        # Add build-system requires if present
        build_deps = self.toml.get("build-system", {}).get("requires", [])
        for dep in build_deps:
            try:
                req = Requirement(dep)
                if req not in all_deps:
                    all_deps.add(req)
            except ValueError:
                continue

        return all_deps

    @property
    def package_name(self) -> str:
        """Get the package name from pyproject.toml"""
        return self.metadata.name

    @property
    def package_version(self) -> Optional[Version]:
        return self.metadata.version

    def get_hwh_config(self) -> HwhConfig:
        return HwhConfig(self.toml)



import os
import shutil
from pathlib import Path
from typing import List, Optional, Union

from Cython.Build import cythonize
from setuptools.build_meta import (
    build_editable as _build_editable,
)
from setuptools.build_meta import (
    get_requires_for_build_editable as _get_requires_for_build_editable,
)
from setuptools.build_meta import (
    get_requires_for_build_wheel as _get_requires_for_build_wheel,
)
from setuptools.build_meta import (
    prepare_metadata_for_build_wheel as _prepare_metadata_for_build_wheel,
)
from setuptools.command.build_ext import build_ext
from setuptools.dist import Distribution
from setuptools.extension import Extension

from .parser import PyProject

# Global flag to prevent double builds
_EXTENSIONS_BUILT = False


def list_files_walk(start_path="."):
    for root, dirs, files in os.walk(start_path):
        for file in files:
            print(os.path.join(root, file))


def find_cython_files(
    source_dir: Path,
    sources: Optional[List[Union[str, Path]]] = None,
    exclude_dirs: Optional[List[str]] = None,
) -> List[Path]:
    if sources:
        # Convert all sources to Path objects relative to source_dir
        res = [
            source_dir / src if isinstance(src, str) else src
            for src in sources
            if str(src).endswith(".pyx")
        ]
        return res

    all_pyx = list(source_dir.rglob("*.pyx"))

    # Filter out all directories that are excluded
    exclude_dirs = set(exclude_dirs or [])
    if exclude_dirs:
        # Convert exclude_dirs to full paths relative to source_dir
        exclude_paths = {source_dir / excluded for excluded in exclude_dirs}
        return [
            pyx
            for pyx in all_pyx
            if not any(exclude_path in pyx.parents for exclude_path in exclude_paths)
        ]

    return list(all_pyx)


def _get_ext_modules(project: PyProject):
    """Get Cython extension modules configuration."""
    import site

    # Get site-packages directory
    site_packages = [site.getsitepackages()[0]]

    # Create directory lists for Extension ctor and cythonize()
    config = project.get_hwh_config().cython
    library_dirs = config.library_dirs + site_packages
    runtime_library_dirs = config.runtime_library_dirs + site_packages
    include_dirs = config.include_dirs + site_packages

    # Find all .pyx files in the package directory
    package_dir = Path(project.package_name)
    pyx_files = find_cython_files(
        package_dir, sources=config.sources, exclude_dirs=config.exclude_dirs
    )

    # Create Extensions
    ext_modules = []
    for pyx_file in pyx_files:
        module_path = (
            str(pyx_file).replace("/", ".").replace("\\", ".").replace(".pyx", "")
        )
        ext_modules.append(
            Extension(
                module_path,
                [str(pyx_file)],
                language=config.language,
                library_dirs=library_dirs,
                runtime_library_dirs=runtime_library_dirs,
            )
        )

    return cythonize(
        ext_modules,
        nthreads=config.nthreads,
        force=config.force,
        annotate=config.annotate,
        compiler_directives=config.compiler_directives.as_dict(),
        include_path=include_dirs,  # This helps find .pxd files
    )


def _build_extension():
    """Build the extension modules."""
    global _EXTENSIONS_BUILT

    if _EXTENSIONS_BUILT:
        return

    project = PyProject(Path())

    name = project.package_name

    dist = Distribution({"name": name, "ext_modules": _get_ext_modules(project)})

    dist.has_ext_modules = lambda: True

    cmd = build_ext(dist)
    cmd.ensure_finalized()
    cmd.run()

    # Copy built extensions to source directory
    built_files = cmd.get_outputs()
    for output in built_files:
        if os.path.exists(output):
            filename = os.path.basename(output)
            target_dir = Path(name)
            target_path = target_dir / filename
            print(f"Copying {output} to {target_path}")
            shutil.copy2(output, target_path)


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    """Build wheel."""
    _build_extension()

    project = PyProject(Path())
    name = project.package_name

    # Configure the distribution
    dist = Distribution(
        {
            "name": name,
            "version": str(project.package_version),
            "ext_modules": _get_ext_modules(project),
            "packages": [name],
            "package_data": {
                # TODO: Remove pyx files, we don't need them right?
                name: ["*.pxd", "*.so"],
            },
            "include_package_data": True,
        }
    )

    dist.script_name = "youcanapparentlyputanythinghere"
    dist.has_ext_modules = lambda: True

    from wheel.bdist_wheel import bdist_wheel as wheel_command

    class BdistWheelCommand(wheel_command):
        def finalize_options(self):
            super().finalize_options()
            self.root_is_pure = False

        def run(self):
            # TODO: check that this isn't required and remove
            self.run_command("build_ext")
            super().run()

    cmd = BdistWheelCommand(dist)
    cmd.dist_dir = wheel_directory

    cmd.ensure_finalized()
    cmd.run()

    # Find the built wheel in the wheel_directory
    wheel_path = next(Path(wheel_directory).glob("*.whl"))
    return wheel_path.name


def build_editable(wheel_directory, config_settings=None, metadata_directory=None):
    """Build editable wheel."""
    _build_extension()
    result = _build_editable(wheel_directory, config_settings, metadata_directory)
    return result


def get_requires_for_build_wheel(config_settings=None):
    """Get requirements for wheel build."""
    return _get_requires_for_build_wheel(config_settings)


def get_requires_for_build_editable(config_settings=None):
    """Get requirements for editable build."""
    return _get_requires_for_build_editable(config_settings)


def prepare_metadata_for_build_wheel(metadata_directory, config_settings=None):
    """Prepare wheel metadata."""
    return _prepare_metadata_for_build_wheel(metadata_directory, config_settings)


def prepare_metadata_for_build_editable(metadata_directory, config_settings=None):
    """Prepare editable wheel metadata."""
    return _prepare_metadata_for_build_wheel(metadata_directory, config_settings)


def build_sdist(sdist_directory, config_settings=None):
    """Build a source distribution."""
    from setuptools.build_meta import build_sdist as _build_sdist

    return _build_sdist(sdist_directory, config_settings)



import os
from dataclasses import dataclass, field
from enum import StrEnum
from typing import Union, get_args, get_origin


class Language(StrEnum):
    C = "c"
    CPP = "c++"


@dataclass
class CythonCompilerWarningDirectives:
    # TODO: Unused atm
    undeclared: bool = False
    unreachable: bool = True
    maybe_uninitialized: bool = False
    unused: bool = False
    unused_arg: bool = False
    unused_result: bool = False
    multiple_declarators: bool = True


@dataclass
class CythonCompilerDirectives:
    # see https://cython.readthedocs.io/en/0.29.x/src/userguide/source_files_and_compilation.html#compiler-directives
    # Defaults are documentation's defauls
    language_level: str = "3"
    VALID_LANGUAGE_LEVELS = {"2", "3", "3str"}

    binding: bool = False
    boundscheck: bool = True
    wraparound: bool = True
    initializedcheck: bool = False
    nonecheck: bool = False
    overflowcheck: bool = False
    # TODO: overflowcheck.fold, need to alter the dict..
    # overflowcheck_fold: bool = True
    embedsignature: bool = False
    cdivision: bool = False
    cdivision_warnings: bool = False
    # TODO: always_allow_keywords: bool = ?
    profile: bool = False
    linetrace: bool = False
    infer_types: bool | None = None
    # TODO: c_string_type: ?? = ??
    # c_string_unicoding: ?? = ??
    type_version_tag: bool = True
    # TODO: unraisable_traceback: bool = ?
    # TODO: iterable_coroutine: bool = ?

    # TODO: add warning directives

    def as_dict(self) -> dict[str, str | bool]:
        """Convert directives to dictionary for cythonize()."""
        return {
            key: value
            for key, value in self.__dict__.items()
            if value is not None  # Skip None values
        }

    def __post_init__(self):
        """Validate types and values after initialization."""
        for field_name, field_value in self.__dict__.items():
            field_type = self.__annotations__[field_name]

            # Type validation
            if get_origin(field_type) is Union:
                allowed_types = get_args(field_type)
                if field_value is not None and not isinstance(
                    field_value, allowed_types[0]
                ):
                    raise TypeError(
                        f"Field {field_name} must be {allowed_types[0].__name__} or None, "
                        f"got {type(field_value).__name__}"
                    )
            else:
                if not isinstance(field_value, field_type):
                    raise TypeError(
                        f"Field {field_name} must be {field_type.__name__}, "
                        f"got {type(field_value).__name__}"
                    )

            # Value validation for specific fields
            if (
                field_name == "language_level"
                and field_value not in self.VALID_LANGUAGE_LEVELS
            ):
                raise ValueError(
                    f"language_level must be one of {self.VALID_LANGUAGE_LEVELS}, "
                    f"got '{field_value}'"
                )


@dataclass
class CythonConfig:
    language: Language = field(default=Language.C)
    compiler_directives: CythonCompilerDirectives = field(
        default_factory=CythonCompilerDirectives
    )
    nthreads: int = field(default_factory=lambda: os.cpu_count() or 1)
    force: bool = False
    annotate: bool = False
    sources: list[str] = field(default_factory=list)
    exclude_dirs: list[str] = field(default_factory=list)
    include_dirs: list[str] = field(default_factory=list)
    library_dirs: list[str] = field(default_factory=list)
    runtime_library_dirs: list[str] = field(default_factory=list)

    def __post_init__(self):
        if isinstance(self.compiler_directives, dict):
            self.compiler_directives = CythonCompilerDirectives(
                **self.compiler_directives
            )

        if isinstance(self.language, str):
            try:
                self.language = Language(self.language.lower())
            except ValueError:
                raise ValueError(
                    f"Invalid language value: {self.language}. Must be one of {[lang.value for lang in Language]}"
                )

    @classmethod
    def from_pyproject(cls, tool_config: dict) -> "CythonConfig":
        cython_config = tool_config.get("cython", {})

        modules = {}
        modules = cython_config.get("modules", {})
        sources = modules.get("sources", [])
        exclude_dirs = modules.get("exclude_dirs", [])
        include_dirs = modules.get("include_dirs", [])
        runtime_library_dirs = modules.get("runtime_library_dirs", [])
        library_dirs = modules.get("library_dirs", [])
        return cls(
            language=cython_config.get("language") or Language.C,
            compiler_directives=CythonCompilerDirectives(
                **cython_config.get("compiler_directives", {})
            ),
            nthreads=cython_config.get("nthreads", os.cpu_count() or 1),
            force=cython_config.get("force", False),
            annotate=cython_config.get("annotate", False),
            sources=sources,
            exclude_dirs=exclude_dirs,
            include_dirs=include_dirs,
            library_dirs=library_dirs,
            runtime_library_dirs=runtime_library_dirs,
        )


class HwhConfig:
    def __init__(self, pyproject_data: dict):
        all_tools = pyproject_data.get("tool")
        config = {}

        if all_tools:
            config = all_tools.get("hwh", {})
        self.cython = CythonConfig.from_pyproject(config)
